from __future__ import annotations

__all__ = ["ConvolutionalGatingMLPV1Config", "ConvolutionalGatingMLPV1"]

from dataclasses import dataclass
from typing import Callable

import torch
from torch import nn

from i6_models.config import ModelConfiguration


@dataclass
class ConvolutionalGatingMLPV1Config(ModelConfiguration):
    """
    Attributes:
        input_dim: input dimension
        hidden_dim: hidden dimension (normally set to 6*input_dim as suggested by the paper)
        kernel_size: kernel size of the depthwise convolution layer
        dropout: dropout probability
        activation: activation function
    """

    input_dim: int
    hidden_dim: int
    kernel_size: int
    dropout: float
    activation: Callable[[torch.Tensor], torch.Tensor] = nn.functional.gelu

    def check_valid(self):
        assert self.kernel_size % 2 == 1, "ConvolutionalGatingMLPV1 only supports odd kernel sizes"
        assert self.hidden_dim % 2 == 0, "ConvolutionalGatingMLPV1 only supports even hidden_dim"

    def __post__init__(self):
        super().__post_init__()
        self.check_valid()


class ConvolutionalGatingMLPV1(nn.Module):
    """Convolutional Gating MLP (cgMLP)."""

    def __init__(self, model_cfg: ConvolutionalGatingMLPV1Config):
        super().__init__()

        self.layer_norm_input = nn.LayerNorm(model_cfg.input_dim)
        self.linear_ff = nn.Linear(in_features=model_cfg.input_dim, out_features=model_cfg.hidden_dim, bias=True)
        self.activation = model_cfg.activation
        self.layer_norm_csgu = nn.LayerNorm(model_cfg.hidden_dim // 2)
        self.depthwise_conv = nn.Conv1d(
            in_channels=model_cfg.hidden_dim // 2,
            out_channels=model_cfg.hidden_dim // 2,
            kernel_size=model_cfg.kernel_size,
            padding=(model_cfg.kernel_size - 1) // 2,
            groups=model_cfg.hidden_dim // 2,
        )
        self.linear_out = nn.Linear(in_features=model_cfg.hidden_dim // 2, out_features=model_cfg.input_dim, bias=True)
        self.dropout = model_cfg.dropout

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        :param x: shape [B, T, F], F=input_dim
        :return: shape [B, T, F], F=input_dim
        """
        x = self.layer_norm_input(x)  # [B, T, F]
        x = self.linear_ff(x)  # [B, T, F']
        x = self.activation(x)

        # convolutional spatial gating unit (csgu)
        x_1, x_2 = x.chunk(2, dim=-1)  # [B, T, F'//2], [B, T, F'//2]
        x_2 = self.layer_norm_csgu(x_2)
        # conv layers expect shape [B, F, T] so we have to transpose here
        x_2 = x_2.transpose(1, 2)  # [B, F'//2, T]
        x_2 = self.depthwise_conv(x_2)  # [B, F'//2, T]
        x_2 = x_2.transpose(1, 2)  # [B, T, F'//2]
        x = x_1 * x_2  # [B, T, F'//2]
        x = nn.functional.dropout(x, p=self.dropout, training=self.training)

        x = self.linear_out(x)  # [B, T, F]
        x = nn.functional.dropout(x, p=self.dropout, training=self.training)
        return x



import numpy as np
import torch


def _mask(tensor: torch.Tensor, batch_axis: int, axis: int, pos: torch.Tensor, max_len: int) -> torch.Tensor:
    """
    :param tensor: e.g. [B, ..., A, ...] but arbitrary axis order
    :param batch_axis: index of the batch axis
    :param axis: which axis A to mask
    :param pos: at which positions along axis to start the mask (size [B])
    :param max_len: mask length drawn uniformly from [0, max_len]
    """
    batch_dim_size = tensor.shape[batch_axis]
    mask_dim_size = tensor.shape[axis]
    mask_len = torch.randint(low=1, high=max_len + 1, size=(batch_dim_size,), dtype=torch.int32, device=tensor.device)
    end_pos = torch.min(pos + mask_len, torch.tensor([mask_dim_size] * batch_dim_size, device=tensor.device))
    idxs = torch.arange(0, mask_dim_size, device=tensor.device).unsqueeze(0)  # [1,dim]
    pos_bc = pos.unsqueeze(1)  # [B,1]
    end_pos_bc = end_pos.unsqueeze(1)  # [B,1]
    mask = torch.logical_and(torch.greater_equal(idxs, pos_bc), torch.less(idxs, end_pos_bc))  # [B,dim]
    if batch_axis > axis:
        mask = mask.transpose(0, 1)  # [dim,B]
    mask = torch.reshape(mask, shape=[tensor.shape[i] if i in (batch_axis, axis) else 1 for i in range(tensor.ndim)])
    tensor = torch.where(mask, 0.0, tensor)
    return tensor


def _random_mask(tensor: torch.Tensor, batch_axis: int, axis: int, min_num: int, max_num: int, max_len: int):
    """
    Mask tensor along axis using N in [min_num, max_num] masks of length [0, max_len]

    :param tensor: e.g. [B, ..., A, ...] but arbitrary axis order
    :param batch_axis: index of the batch axis
    :param axis: which axis to mask
    :param min_num: minimum number of masks
    :param max_num: maximum number of masks
    :param max_amount: mask length drawn uniformly from [0, max_amount]
    """

    batch_dim_size = tensor.shape[batch_axis]
    if max_num < min_num:
        max_num = min_num
    num_masks = torch.randint(min_num, max_num + 1, size=(batch_dim_size,), device="cpu")  # [B]

    max_num_masks = num_masks.max().item()

    z = torch.rand((batch_dim_size, tensor.shape[axis]), device=tensor.device)  # [B,dim]
    _, indices = torch.topk(z, max_num_masks, dim=1)

    # Make num_masks broadcastable to shape of tensor for torch.where.
    num_masks = torch.reshape(num_masks, [1] * batch_axis + [batch_dim_size] + [1] * (tensor.dim() - batch_axis - 1))

    num_masks = num_masks.to(device=tensor.device)

    for i in range(max_num_masks):
        tensor = torch.where(i < num_masks, _mask(tensor, batch_axis, axis, indices[:, i], max_len), tensor)

    return tensor


def specaugment_v1(
    audio_features: torch.Tensor,
    *,
    time_min_num_masks: int,
    time_max_num_masks: int,
    time_mask_max_size: int,
    freq_min_num_masks: int,
    freq_max_num_masks: int,
    freq_mask_max_size: int,
):
    """
    Specaugment from legacy rossenbach/zeineldeen/zeyer attention setups e.g.,
    https://github.com/rwth-i6/i6_experiments/blob/main/users/zeineldeen/data_aug/specaugment/specaug_tf2.py
    but without any step-based scheduling and without dependence on length.
    See `specaugment_v1_by_length` for a variant which is more close to the original.

    Fills masks with zeros.

    Basically just a convenience wrapper around _random_mask.

    See also: https://arxiv.org/abs/1904.08779

    :param audio_features: e.g. log-mel features as [B, T, F]
    :param time_min_num_masks: minimum number of masks along T
    :param time_max_num_masks: maximum number of masks along T
    :param time_mask_max_size: maximum size of masks along T
    :param freq_min_num_masks: minimum number of masks along F
    :param freq_max_num_masks: maximum number of masks along F
    :param freq_mask_max_size: maximum size of masks along F
    :return: masked audio features
    """
    assert len(audio_features.shape) == 3
    assert time_min_num_masks <= time_max_num_masks
    assert freq_min_num_masks <= freq_max_num_masks
    masked_audio_features = _random_mask(
        audio_features, 0, 1, time_min_num_masks, time_max_num_masks, time_mask_max_size
    )  # time masking
    masked_audio_features = _random_mask(
        masked_audio_features, 0, 2, freq_min_num_masks, freq_max_num_masks, freq_mask_max_size
    )  # freq masking
    return masked_audio_features


def specaugment_v1_by_length(
    audio_features: torch.Tensor,
    *,
    time_min_num_masks: int,
    time_max_mask_per_n_frames: int,
    time_mask_max_size: int,
    freq_min_num_masks: int,
    freq_max_num_masks: int,
    freq_mask_max_size: int,
):
    """
    Convenience wrapper around specaugment_v1 with time-length adaptive number of masks.

    :param audio_features: e.g. log-mel features as [B, T, F]
    :param time_max_mask_per_n_frames: used for the maximum number time masks,
        max_num_masks = T / max_mask_per_n_frames for each batch.
        They are still drawn depending on the full batch length, so shorter sequences
        might get more masks than that by chance, or none at all when all masks
        fall into the padding space.
    :param time_min_num_masks: minimum number of masks along T
    :param time_mask_max_size: maximum size of masks along T
    :param freq_min_num_masks: minimum number of masks along F
    :param freq_max_num_masks: maximum number of masks along F
    :param freq_mask_max_size: maximum size of masks along F
    :return: masked audio features
    """
    return specaugment_v1(
        audio_features,
        time_min_num_masks=time_min_num_masks,
        time_max_num_masks=np.maximum(audio_features.size(1) // time_max_mask_per_n_frames, time_min_num_masks),
        time_mask_max_size=time_mask_max_size,
        freq_min_num_masks=freq_min_num_masks,
        freq_max_num_masks=freq_max_num_masks,
        freq_mask_max_size=freq_mask_max_size,
    )



__all__ = ["MultiheadSelfAttentionV1", "MultiheadSelfAttentionV1Config"]

import math
from dataclasses import dataclass
import torch

from i6_models.config import ModelConfiguration
from i6_models.util import compat


@dataclass
class MultiheadSelfAttentionV1Config(ModelConfiguration):
    """
    Attributes:
        input_dim: input dim and total dimension for query/key and value projections, should be divisible by `num_att_heads`
        num_att_heads: number of attention heads
        att_weights_dropout: attention weights dropout
        dropout: attention weight dropout probability
    """

    input_dim: int
    num_att_heads: int
    att_weights_dropout: float
    dropout: float

    def __post_init__(self) -> None:
        super().__post_init__()
        assert self.input_dim % self.num_att_heads == 0, "input_dim must be divisible by num_att_heads"


class MultiheadSelfAttentionV1(torch.nn.Module):
    """
    Native Multihead Self Attention implementation based on 'Attention Is All You Need'
    """

    def __init__(self, cfg: MultiheadSelfAttentionV1Config):
        super().__init__()
        self.cfg = cfg
        self.num_att_heads = cfg.num_att_heads
        self.input_dim = cfg.input_dim
        self.dim_heads = self.input_dim // self.num_att_heads
        self.layernorm = torch.nn.LayerNorm(cfg.input_dim)

        self.out_proj = torch.nn.Linear(in_features=cfg.input_dim, out_features=cfg.input_dim, bias=True)
        self.in_proj = torch.nn.Linear(in_features=cfg.input_dim, out_features=3 * cfg.input_dim, bias=True)

        self.norm = math.sqrt(float(self.input_dim / self.num_att_heads))
        self.softmax = torch.nn.Softmax(-1)
        self.dropout = torch.nn.Dropout(cfg.att_weights_dropout)

    def forward(self, qkv: torch.Tensor, key_padding_mask: torch.Tensor):
        """
        Computes the forward pass of the MultiheadSelfAttentionV1 module.
        Attributes:

            qkv (torch.Tensor): The input tensor of shape (B, T, F).
            key_padding_mask (torch.Tensor): The key padding mask tensor of shape (batch_dim, num_tokens).
        """

        batch_dim, num_tokens, embed_dim = qkv.shape
        x = self.in_proj(qkv)

        hidden_dim = qkv.size(-1)
        query, key, value = x.unflatten(-1, (3, hidden_dim)).unsqueeze(0).transpose(0, -2).squeeze(-2).contiguous()

        query = query.view(batch_dim, -1, self.num_att_heads, self.dim_heads)  # [B, T, D//H, D']
        key = key.view(batch_dim, -1, self.num_att_heads, self.dim_heads)  # [B, T, D//H, D']
        value = value.view(batch_dim, -1, self.num_att_heads, self.dim_heads)  # [B, T, D//H, D']

        query = torch.transpose(query, 1, 2)  # [B, D//H, T, D']
        key = torch.transpose(key, 1, 2)  # [B, D//H, T, D']
        value = torch.transpose(value, 1, 2)  # [B, D//H, T, D']

        key = torch.transpose(key, -2, -1)  # [B, D//H, D', T]

        dot = torch.matmul(query, key)  # [B, D//H, T, T]
        dot = dot / self.norm

        if key_padding_mask is not None:
            key_padding_mask = key_padding_mask.view(batch_dim, 1, 1, key_padding_mask.size(1))
            dot = dot.masked_fill(key_padding_mask, -float("inf"))

        alpha = self.softmax(dot)  # [B, D//H, T, T]
        alpha = self.dropout(alpha)

        att_out = torch.matmul(alpha, value)  # [B, D//H, T, D']
        att_out = torch.transpose(att_out, 1, 2)  # [B, T, D//H, D']
        att_out = att_out.reshape(batch_dim, -1, self.input_dim)  # [B, T, D]
        att_out = self.out_proj(att_out)

        return att_out, alpha



from __future__ import annotations

__all__ = [
    "VGG4LayerActFrontendV1",
    "VGG4LayerActFrontendV1Config",
]

from dataclasses import dataclass
from typing import Callable, Optional, Tuple, Union

import torch
from torch import nn

from i6_models.config import ModelConfiguration

from .common import get_same_padding, mask_pool, calculate_output_dim


@dataclass
class VGG4LayerActFrontendV1Config(ModelConfiguration):
    """
    Attributes:
        in_features: number of input features to module
        conv1_channels: number of channels for first conv layer
        conv2_channels: number of channels for second conv layer
        conv3_channels: number of channels for third conv layer
        conv4_channels: number of channels for fourth conv layer
        conv_kernel_size: kernel size of conv layers
        conv_padding: padding for the convolution
        pool1_kernel_size: kernel size of first pooling layer
        pool1_stride: stride of first pooling layer
        pool1_padding: padding for first pooling layer
        pool2_kernel_size: kernel size of second pooling layer
        pool2_stride: stride of second pooling layer
        pool2_padding: padding for second pooling layer
        activation: activation function at the end
        out_features: output size of the final linear layer
    """

    in_features: int
    conv1_channels: int
    conv2_channels: int
    conv3_channels: int
    conv4_channels: int
    conv_kernel_size: Tuple[int, int]
    conv_padding: Optional[Tuple[int, int]]
    pool1_kernel_size: Tuple[int, int]
    pool1_stride: Optional[Tuple[int, int]]
    pool1_padding: Optional[Tuple[int, int]]
    pool2_kernel_size: Tuple[int, int]
    pool2_stride: Optional[Tuple[int, int]]
    pool2_padding: Optional[Tuple[int, int]]
    activation: Union[nn.Module, Callable[[torch.Tensor], torch.Tensor]]
    out_features: int

    def check_valid(self):
        if isinstance(self.conv_kernel_size, int):
            assert self.conv_kernel_size % 2 == 1, "ConformerVGGFrontendV1 only supports odd kernel sizes"
        if isinstance(self.pool1_kernel_size, int):
            assert self.pool1_kernel_size % 2 == 1, "ConformerVGGFrontendV1 only supports odd kernel sizes"
        if isinstance(self.pool2_kernel_size, int):
            assert self.pool2_kernel_size % 2 == 1, "ConformerVGGFrontendV1 only supports odd kernel sizes"

    def __post__init__(self):
        super().__post_init__()
        self.check_valid()


class VGG4LayerActFrontendV1(nn.Module):
    """
    Convolutional Front-End

    The frond-end utilizes convolutional and pooling layers, as well as activation functions
    to transform a feature vector, typically Log-Mel or Gammatone for audio, into an intermediate
    representation.

    Structure of the front-end:
      - Conv
      - Conv
      - Activation
      - Pool
      - Conv
      - Conv
      - Activation
      - Pool

    Uses explicit padding for ONNX exportability, see:
    https://github.com/pytorch/pytorch/issues/68880
    """

    def __init__(self, model_cfg: VGG4LayerActFrontendV1Config):
        """
        :param model_cfg: model configuration for this module
        """
        super().__init__()

        model_cfg.check_valid()

        self.cfg = model_cfg

        conv_padding = (
            model_cfg.conv_padding
            if model_cfg.conv_padding is not None
            else get_same_padding(model_cfg.conv_kernel_size)
        )
        pool1_padding = model_cfg.pool1_padding if model_cfg.pool1_padding is not None else (0, 0)
        pool2_padding = model_cfg.pool2_padding if model_cfg.pool2_padding is not None else (0, 0)

        self.conv1 = nn.Conv2d(
            in_channels=1,
            out_channels=model_cfg.conv1_channels,
            kernel_size=model_cfg.conv_kernel_size,
            padding=conv_padding,
        )
        self.conv2 = nn.Conv2d(
            in_channels=model_cfg.conv1_channels,
            out_channels=model_cfg.conv2_channels,
            kernel_size=model_cfg.conv_kernel_size,
            padding=conv_padding,
        )
        self.pool1 = nn.MaxPool2d(
            kernel_size=model_cfg.pool1_kernel_size,
            stride=model_cfg.pool1_stride,
            padding=pool1_padding,
        )
        self.conv3 = nn.Conv2d(
            in_channels=model_cfg.conv2_channels,
            out_channels=model_cfg.conv3_channels,
            kernel_size=model_cfg.conv_kernel_size,
            padding=conv_padding,
        )
        self.conv4 = nn.Conv2d(
            in_channels=model_cfg.conv3_channels,
            out_channels=model_cfg.conv4_channels,
            kernel_size=model_cfg.conv_kernel_size,
            padding=conv_padding,
        )
        self.pool2 = nn.MaxPool2d(
            kernel_size=model_cfg.pool2_kernel_size,
            stride=model_cfg.pool2_stride,
            padding=pool2_padding,
        )
        self.activation = model_cfg.activation
        self.linear = nn.Linear(
            in_features=self._calculate_dim(),
            out_features=model_cfg.out_features,
            bias=True,
        )

    def forward(self, tensor: torch.Tensor, sequence_mask: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        T might be reduced to T' or T'' depending on stride of the layers

        stride is only allowed for the pool1 and pool2 operation.
        other ops do not have stride configurable -> no update of mask sequence required but added anyway

        :param tensor: input tensor of shape [B,T,F]
        :param sequence_mask: the sequence mask for the tensor
        :return: torch.Tensor of shape [B,T",F'] and the shape of the sequence mask
        """
        assert tensor.shape[-1] == self.cfg.in_features, f"shape {tensor.shape} vs in features {self.cfg.in_features}"
        # and add a dim
        tensor = tensor[:, None, :, :]  # [B,C=1,T,F]

        tensor = self.conv1(tensor)
        sequence_mask = mask_pool(
            seq_mask=sequence_mask,
            kernel_size=self.conv1.kernel_size[0],
            stride=self.conv1.stride[0],
            padding=self.conv1.padding[0],
        )

        tensor = self.conv2(tensor)
        sequence_mask = mask_pool(
            sequence_mask,
            kernel_size=self.conv2.kernel_size[0],
            stride=self.conv2.stride[0],
            padding=self.conv2.padding[0],
        )

        tensor = self.activation(tensor)
        tensor = self.pool1(tensor)  # [B,C,T',F']
        sequence_mask = mask_pool(
            sequence_mask,
            kernel_size=self.pool1.kernel_size[0],
            stride=self.pool1.stride[0],
            padding=self.pool1.padding[0],
        )

        tensor = self.conv3(tensor)
        sequence_mask = mask_pool(
            sequence_mask,
            kernel_size=self.conv3.kernel_size[0],
            stride=self.conv3.stride[0],
            padding=self.conv3.padding[0],
        )

        tensor = self.conv4(tensor)
        sequence_mask = mask_pool(
            sequence_mask,
            kernel_size=self.conv4.kernel_size[0],
            stride=self.conv4.stride[0],
            padding=self.conv4.padding[0],
        )

        tensor = self.activation(tensor)
        tensor = self.pool2(tensor)  # [B,C,T",F"]
        sequence_mask = mask_pool(
            sequence_mask,
            kernel_size=self.pool2.kernel_size[0],
            stride=self.pool2.stride[0],
            padding=self.pool2.padding[0],
        )

        tensor = torch.transpose(tensor, 1, 2)  # transpose to [B,T",C,F"]
        tensor = torch.flatten(tensor, start_dim=2, end_dim=-1)  # [B,T",C*F"]

        tensor = self.linear(tensor)

        return tensor, sequence_mask

    def _calculate_dim(self) -> int:
        # conv1
        out_dim = calculate_output_dim(
            in_dim=self.cfg.in_features,
            filter_size=self.conv1.kernel_size[1],
            stride=self.conv1.stride[1],
            padding=self.conv1.padding[1],
        )
        # conv2
        out_dim = calculate_output_dim(
            in_dim=out_dim,
            filter_size=self.conv2.kernel_size[1],
            stride=self.conv2.stride[1],
            padding=self.conv2.padding[1],
        )
        # pool1
        out_dim = calculate_output_dim(
            in_dim=out_dim,
            filter_size=self.pool1.kernel_size[1],
            stride=self.pool1.stride[1],
            padding=self.pool1.padding[1],
        )
        # conv3
        out_dim = calculate_output_dim(
            in_dim=out_dim,
            filter_size=self.conv3.kernel_size[1],
            stride=self.conv3.stride[1],
            padding=self.conv3.padding[1],
        )
        # conv4
        out_dim = calculate_output_dim(
            in_dim=out_dim,
            filter_size=self.conv4.kernel_size[1],
            stride=self.conv4.stride[1],
            padding=self.conv4.padding[1],
        )
        # pool2
        out_dim = calculate_output_dim(
            in_dim=out_dim,
            filter_size=self.pool2.kernel_size[1],
            stride=self.pool2.stride[1],
            padding=self.pool2.padding[1],
        )
        out_dim *= self.conv4.out_channels
        return out_dim



from __future__ import annotations

__all__ = ["ConformerPositionwiseFeedForwardV1", "ConformerPositionwiseFeedForwardV1Config"]

from dataclasses import dataclass
from typing import Callable

import torch
from torch import nn

from i6_models.config import ModelConfiguration


@dataclass
class ConformerPositionwiseFeedForwardV1Config(ModelConfiguration):
    """
    Attributes:
        input_dim: input dimension
        hidden_dim: hidden dimension (normally set to 4*input_dim as suggested by the paper)
        dropout: dropout probability
        activation: activation function
    """

    input_dim: int
    hidden_dim: int
    dropout: float
    activation: Callable[[torch.Tensor], torch.Tensor] = nn.functional.silu


class ConformerPositionwiseFeedForwardV1(nn.Module):
    """
    Conformer feedforward module
    """

    def __init__(self, cfg: ConformerPositionwiseFeedForwardV1Config):
        super().__init__()

        self.layer_norm = nn.LayerNorm(cfg.input_dim)
        self.linear_ff = nn.Linear(in_features=cfg.input_dim, out_features=cfg.hidden_dim, bias=True)
        self.activation = cfg.activation
        self.linear_out = nn.Linear(in_features=cfg.hidden_dim, out_features=cfg.input_dim, bias=True)
        self.dropout = cfg.dropout

    def forward(self, tensor: torch.Tensor) -> torch.Tensor:
        """
        :param tensor: shape [B,T,F], F=input_dim
        :return: shape [B,T,F], F=input_dim
        """
        tensor = self.layer_norm(tensor)
        tensor = self.linear_ff(tensor)  # [B,T,F]
        tensor = self.activation(tensor)  # [B,T,F]
        tensor = nn.functional.dropout(tensor, p=self.dropout, training=self.training)  # [B,T,F]
        tensor = self.linear_out(tensor)  # [B,T,F]
        tensor = nn.functional.dropout(tensor, p=self.dropout, training=self.training)  # [B,T,F]
        return tensor



__all__ = [
    "LogMelFeatureExtractionV1",
    "LogMelFeatureExtractionV1Config",
    "RasrCompatibleLogMelFeatureExtractionV1",
    "RasrCompatibleLogMelFeatureExtractionV1Config",
]

import math
from dataclasses import dataclass
from typing import Optional, Tuple

from librosa import filters
import torch
from torch import nn

from i6_models.config import ModelConfiguration


@dataclass
class LogMelFeatureExtractionV1Config(ModelConfiguration):
    """
    Attributes:
        sample_rate: audio sample rate in Hz
        win_size: window size in seconds
        hop_size: window shift in seconds
        f_min: minimum filter frequency in Hz
        f_max: maximum filter frequency in Hz
        min_amp: minimum amplitude for safe log
        num_filters: number of mel windows
        center: centered STFT with automatic padding
    """

    sample_rate: int
    win_size: float
    hop_size: float
    f_min: int
    f_max: int
    min_amp: float
    num_filters: int
    center: bool
    n_fft: Optional[int] = None

    def __post_init__(self) -> None:
        super().__post_init__()
        assert self.f_max <= self.sample_rate // 2, "f_max can not be larger than half of the sample rate"
        assert self.f_min >= 0 and self.f_max > 0 and self.sample_rate > 0, "frequencies need to be positive"
        assert self.win_size > 0 and self.hop_size > 0, "window settings need to be positive"
        assert self.num_filters > 0, "number of filters needs to be positive"
        assert self.hop_size <= self.win_size, "using a larger hop size than window size does not make sense"
        if self.n_fft is None:
            # if n_fft is not given, set n_fft to the window size (in samples)
            self.n_fft = int(self.win_size * self.sample_rate)
        else:
            assert self.n_fft >= self.win_size * self.sample_rate, "n_fft cannot to be smaller than the window size"


class LogMelFeatureExtractionV1(nn.Module):
    """
    Librosa-compatible log-mel feature extraction using log10. Does not use torchaudio.

    Using it wrapped with torch.no_grad() is recommended if no gradient is needed
    """

    def __init__(self, cfg: LogMelFeatureExtractionV1Config):
        super().__init__()
        self.center = cfg.center
        self.hop_length = int(cfg.hop_size * cfg.sample_rate)
        self.min_amp = cfg.min_amp
        self.n_fft = cfg.n_fft
        self.win_length = int(cfg.win_size * cfg.sample_rate)

        self.register_buffer(
            "mel_basis",
            torch.tensor(
                filters.mel(
                    sr=cfg.sample_rate,
                    n_fft=cfg.n_fft,
                    n_mels=cfg.num_filters,
                    fmin=cfg.f_min,
                    fmax=cfg.f_max,
                )
            ),
        )
        self.register_buffer("window", torch.hann_window(self.win_length))

    def forward(self, raw_audio, length) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        :param raw_audio: [B, T]
        :param length in samples: [B]
        :return features as [B,T,F] and length in frames [B]
        """
        power_spectrum = (
            torch.abs(
                torch.stft(
                    raw_audio,
                    n_fft=self.n_fft,
                    hop_length=self.hop_length,
                    win_length=self.win_length,
                    window=self.window,
                    center=self.center,
                    pad_mode="constant",
                    return_complex=True,
                )
            )
            ** 2
        )
        if len(power_spectrum.size()) == 2:
            # For some reason torch.stft removes the batch axis for batch sizes of 1, so we need to add it again
            power_spectrum = torch.unsqueeze(power_spectrum, 0)
        melspec = torch.einsum("...ft,mf->...mt", power_spectrum, self.mel_basis)
        log_melspec = torch.log10(torch.clamp(melspec, min=self.min_amp))
        feature_data = torch.transpose(log_melspec, 1, 2)

        if self.center:
            length = (length // self.hop_length) + 1
        else:
            length = ((length - self.n_fft) // self.hop_length) + 1

        return feature_data, length.int()


@dataclass
class RasrCompatibleLogMelFeatureExtractionV1Config(ModelConfiguration):
    """
    Attributes:
        sample_rate: audio sample rate in Hz
        win_size: window size in seconds
        hop_size: window shift in seconds
        min_amp: minimum amplitude for safe log
        num_filters: number of mel windows
        alpha: preemphasis weight
    """

    sample_rate: int
    win_size: float
    hop_size: float
    min_amp: float
    num_filters: int
    alpha: float = 1.0

    def __post_init__(self) -> None:
        super().__post_init__()
        assert self.win_size > 0 and self.hop_size > 0, "window settings need to be positive"
        assert self.num_filters > 0, "number of filters needs to be positive"
        assert self.hop_size <= self.win_size, "using a larger hop size than window size does not make sense"


class RasrCompatibleLogMelFeatureExtractionV1(nn.Module):
    """
    Rasr-compatible log-mel feature extraction using log10. Does not use torchaudio.
    """

    def __init__(self, cfg: RasrCompatibleLogMelFeatureExtractionV1Config):
        super().__init__()

        self.sample_rate = int(cfg.sample_rate)
        self.hop_length = int(cfg.hop_size * cfg.sample_rate)
        self.min_amp = cfg.min_amp
        self.win_length = int(cfg.win_size * cfg.sample_rate)
        self.n_fft = 2 ** math.ceil(
            math.log2(self.win_length)
        )  # smallest power if two which is greater than or equal to win_length
        self.alpha = cfg.alpha

        self.register_buffer(
            "mel_basis",
            torch.tensor(
                filters.mel(
                    sr=cfg.sample_rate,
                    n_fft=self.n_fft,
                    n_mels=cfg.num_filters,
                    fmin=0,
                    fmax=cfg.sample_rate // 2,
                    htk=True,
                    norm=None,
                ),
            ),
        )
        self.register_buffer(
            "window", torch.hann_window(self.win_length, periodic=False, dtype=torch.float64).to(torch.float32)
        )

    def forward(self, raw_audio, length) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        :param raw_audio: [B, T]
        :param length: in samples [B]
        :return features as [B,T,F] and length in frames [B]
        """
        assert raw_audio.shape[1] > 0  # also same for length
        res_size = max(raw_audio.shape[1] - self.win_length + self.hop_length - 1, 0) // self.hop_length + 1
        res_length = (
            torch.maximum(length - self.win_length + self.hop_length - 1, torch.zeros_like(length)) // self.hop_length
            + 1
        )

        # preemphasize
        preemphasized = raw_audio.clone()
        preemphasized[..., 1:] -= self.alpha * preemphasized[..., :-1]
        preemphasized[..., 0] = 0.0

        # zero pad for the last frame of each sequence in the batch
        last_win_size = length - (res_length - 1) * self.hop_length  # [B]
        last_pad = self.win_length - last_win_size  # [B]

        # zero pad for the whole batch
        last_pad_batch = self.win_length - (preemphasized.shape[1] - (res_size - 1) * self.hop_length)
        padded = torch.nn.functional.pad(preemphasized, (0, last_pad_batch))

        windowed = padded.unfold(1, size=self.win_length, step=self.hop_length)  # [B, T', W=self.win_length]

        smoothed = windowed * self.window[None, None, :]  # [B, T', W]

        # The last window might be shorter. Will use a shorter Hanning window then. Need to fix that.
        for i, (last_w_size, last_p, res_l) in enumerate(zip(last_win_size, last_pad, res_length)):
            last_win = torch.hann_window(last_w_size, periodic=False, dtype=torch.float64).to(
                self.window.device, torch.float32
            )
            last_win = torch.nn.functional.pad(last_win, (0, last_p))  # [W]
            smoothed[i, res_l - 1] = windowed[i, res_l - 1] * last_win[None, :]

        # compute amplitude spectrum using torch.fft.rfftn with Rasr specific scaling
        fft = torch.fft.rfftn(smoothed, s=self.n_fft) / self.sample_rate  # [B, T', F=n_fft//2+1]
        amplitude_spectrum = torch.abs(fft)  # [B, T', F=n_fft//2+1]

        melspec = torch.einsum("...tf,mf->...tm", amplitude_spectrum, self.mel_basis)  # [B, T', F'=num_filters]
        log_melspec = torch.log10(melspec + self.min_amp)

        return log_melspec, res_length



__all__ = ["FeedForwardEncoderV1Config", "FeedForwardEncoderV1"]

from typing import Tuple
from dataclasses import dataclass
import torch
from torch import nn

from i6_models.parts.ffnn import FeedForwardLayerV1, FeedForwardLayerV1Config
from i6_models.config import ModelConfiguration, ModuleFactoryV1


@dataclass
class FeedForwardEncoderV1Config(ModelConfiguration):
    """
    Attributes:
        num_layers: number of feed-forward layers
        frontend: module factory for the frontend
        layer_cfg: configuration object for each feed-forward layer
    """

    num_layers: int
    frontend: ModuleFactoryV1
    layer_cfg: FeedForwardLayerV1Config


class FeedForwardEncoderV1(nn.Module):
    """
    Simple feed-forward encoder.
    Subsampling can be achieved by setting stride > 1 in the frontend config.
    """

    def __init__(self, cfg: FeedForwardEncoderV1Config):
        super().__init__()
        self.frontend = cfg.frontend()
        self.module_list = nn.ModuleList([FeedForwardLayerV1(cfg.layer_cfg) for _ in range(cfg.num_layers)])

    def forward(self, data_tensor: torch.Tensor, sequence_mask: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        x, sequence_mask = self.frontend(data_tensor, sequence_mask)  # [B, T, F']
        for module in self.module_list:
            x, sequence_mask = module(x, sequence_mask)  # [B, T, F']

        return x, sequence_mask



from __future__ import annotations

__all__ = ["ConformerBlockV2Config", "ConformerEncoderV2Config", "ConformerBlockV2", "ConformerEncoderV2"]

import torch
from torch import nn

from dataclasses import dataclass, field
from typing import Tuple, List, Optional

from i6_models.config import ModelConfiguration, ModuleFactoryV1
from .conformer_v1 import (
    ConformerConvolutionV1,
    ConformerConvolutionV1Config,
    ConformerMHSAV1,
    ConformerMHSAV1Config,
    ConformerPositionwiseFeedForwardV1,
    ConformerPositionwiseFeedForwardV1Config,
)


@dataclass
class ConformerBlockV2Config(ModelConfiguration):
    """
    Attributes:
        ff_cfg: Configuration for ConformerPositionwiseFeedForwardV1
        mhsa_cfg: Configuration for ConformerMHSAV1
        conv_cfg: Configuration for ConformerConvolutionV1
        modules: List of modules to use for ConformerBlockV2,
            "ff" for feed forward module, "mhsa" for multi-head self attention module, "conv" for conv module
        scales: List of scales to apply to the module outputs before the residual connection
    """

    # nested configurations
    ff_cfg: ConformerPositionwiseFeedForwardV1Config
    mhsa_cfg: ConformerMHSAV1Config
    conv_cfg: ConformerConvolutionV1Config
    modules: List[str] = field(default_factory=lambda: ["ff", "mhsa", "conv", "ff"])
    scales: List[float] = field(default_factory=lambda: [0.5, 1.0, 1.0, 0.5])

    def __post__init__(self):
        super().__post_init__()
        assert len(self.modules) == len(self.scales), "modules and scales must have same length"
        for module_name in self.modules:
            assert module_name in ["ff", "mhsa", "conv"], "module not supported"


class ConformerBlockV2(nn.Module):
    """
    Modifications compared to ConformerBlockV1:
    - more generic, enable constructing the block with self-defined module_list.
    """

    def __init__(self, cfg: ConformerBlockV2Config):
        """
        :param cfg: conformer block configuration with subunits for the different conformer parts
        """
        super().__init__()

        modules = []
        for module_name in cfg.modules:
            if module_name == "ff":
                modules.append(ConformerPositionwiseFeedForwardV1(cfg=cfg.ff_cfg))
            elif module_name == "mhsa":
                modules.append(ConformerMHSAV1(cfg=cfg.mhsa_cfg))
            elif module_name == "conv":
                modules.append(ConformerConvolutionV1(model_cfg=cfg.conv_cfg))
            else:
                raise NotImplementedError

        self.module_list = nn.ModuleList(modules)
        self.scales = cfg.scales
        self.final_layer_norm = torch.nn.LayerNorm(cfg.ff_cfg.input_dim)

    def forward(self, x: torch.Tensor, /, sequence_mask: torch.Tensor) -> torch.Tensor:
        """
        :param x: input tensor of shape [B, T, F]
        :param sequence_mask: mask tensor where 0 defines positions within the sequence and 1 outside, shape: [B, T]
        :return: torch.Tensor of shape [B, T, F]
        """
        for scale, module in zip(self.scales, self.module_list):
            if isinstance(module, ConformerMHSAV1):
                x = scale * module(x, sequence_mask) + x
            else:
                x = scale * module(x) + x

        x = self.final_layer_norm(x)  #  [B, T, F]
        return x


@dataclass
class ConformerEncoderV2Config(ModelConfiguration):
    """
    Attributes:
        num_layers: Number of conformer layers in the conformer encoder
        frontend: A pair of ConformerFrontend and corresponding config
        block_cfg: Configuration for ConformerBlockV2
    """

    num_layers: int

    # nested configurations
    frontend: ModuleFactoryV1
    block_cfg: ConformerBlockV2Config


class ConformerEncoderV2(nn.Module):
    """
    Modifications compared to ConformerEncoderV1:
    - modules inside each conformer block could be customized.
    - return a list of layer outputs with desired layer indices
    """

    def __init__(self, cfg: ConformerEncoderV2Config):
        """
        :param cfg: conformer encoder configuration with subunits for frontend and conformer blocks
        """
        super().__init__()

        self.frontend = cfg.frontend()
        self.module_list = torch.nn.ModuleList([ConformerBlockV2(cfg.block_cfg) for _ in range(cfg.num_layers)])

    def forward(
        self, data_tensor: torch.Tensor, /, sequence_mask: torch.Tensor, return_layers: Optional[List[int]] = None
    ) -> Tuple[List[torch.Tensor], torch.Tensor]:
        """
        :param data_tensor: input tensor of shape [B, T', F]
        :param sequence_mask: mask tensor where 1 defines positions within the sequence and 0 outside, shape: [B, T']
        :param return_layers: list of layer indices specifying which layers to return, starting from 0
        :return: (outputs, out_seq_mask)
            where outputs is a list of torch.Tensor of shape [B, T, F']
            for each of the layers in return_layers,
            out_seq_mask is a torch.Tensor of shape [B, T]

        F: input feature dim, F': internal and output feature dim
        T': data time dim, T: down-sampled time dim (internal time dim)
        """

        if return_layers is None:
            return_layers = [len(self.module_list) - 1]

        x, sequence_mask = self.frontend(data_tensor, sequence_mask)  # [B, T, F']

        outputs = []
        assert (
            max(return_layers) < len(self.module_list) and min(return_layers) >= 0
        ), f"invalid layer index, should be between 0 and {len(self.module_list)-1}"

        for i in range(max(return_layers) + 1):
            x = self.module_list[i](x, sequence_mask)  # [B, T, F']
            if i in return_layers:
                outputs.append(x)

        return outputs, sequence_mask



from __future__ import annotations

from dataclasses import dataclass
from typing import List, Optional, Tuple

import torch
from torch import nn
from torch.nn import functional

from i6_models.parts.frontend.vgg_act import (
    VGG4LayerActFrontendV1,
    VGG4LayerActFrontendV1Config,
)


@dataclass
class VGG4LayerActTestParams:
    batch: int
    time: int
    features: int
    conv1_channels: int
    conv2_channels: int
    conv3_channels: int
    conv4_channels: int
    pool1_kernel_size: Tuple[int, int]
    pool1_stride: Optional[Tuple[int, int]]
    pool2_kernel_size: Tuple[int, int]
    pool2_stride: Optional[Tuple[int, int]]
    out_features: Optional[int]
    output_shape: List[int]
    in_sequence_mask: torch.Tensor
    out_sequence_mask: torch.Tensor


def test_conformer_vgg_layer_act_frontend_v1():
    torch.manual_seed(42)

    def get_output_shape(test_parameters: VGG4LayerActTestParams):
        data_input = torch.randn(
            test_parameters.batch,
            test_parameters.time,
            test_parameters.features,
        )

        cfg = VGG4LayerActFrontendV1Config(
            in_features=test_parameters.features,
            conv1_channels=test_parameters.conv1_channels,
            conv2_channels=test_parameters.conv2_channels,
            conv3_channels=test_parameters.conv3_channels,
            conv4_channels=test_parameters.conv4_channels,
            conv_kernel_size=(3, 3),
            conv_padding=None,
            pool1_kernel_size=test_parameters.pool1_kernel_size,
            pool1_stride=test_parameters.pool1_stride,
            pool1_padding=None,
            pool2_kernel_size=test_parameters.pool2_kernel_size,
            pool2_stride=test_parameters.pool2_stride,
            pool2_padding=None,
            activation=nn.functional.relu,
            out_features=test_parameters.out_features,
        )

        output, sequence_mask = VGG4LayerActFrontendV1(cfg)(
            data_input,
            test_parameters.in_sequence_mask,
        )

        return output.shape, sequence_mask

    for idx, test_params in enumerate(
        [
            VGG4LayerActTestParams(  # idx 0
                batch=10,  # int
                time=50,  # int
                features=50,  # int
                conv1_channels=1,  # int
                conv2_channels=1,  # int
                conv3_channels=1,  # int
                conv4_channels=1,  # int
                pool1_kernel_size=(1, 1),  # Tuple[int, int]
                pool1_stride=(1, 1),  # Optional[Tuple[int, int]]
                pool2_kernel_size=(1, 1),  # Tuple[int, int]
                pool2_stride=(1, 1),  # Optional[Tuple[int, int]]
                out_features=50,  # int
                output_shape=[10, 50, 50],  # Tuple[int, int, int]
                in_sequence_mask=torch.Tensor(10 * [25 * [True] + 25 * [False]]).bool(),  # torch.Tensor
                out_sequence_mask=torch.Tensor(10 * [25 * [True] + 25 * [False]]).bool(),  # torch.Tensor
            ),
            VGG4LayerActTestParams(  # idx 1
                batch=10,  # int
                time=50,  # int
                features=50,  # int
                conv1_channels=1,  # int
                conv2_channels=1,  # int
                conv3_channels=1,  # int
                conv4_channels=1,  # int
                pool1_kernel_size=(2, 1),  # Tuple[int, int]
                pool1_stride=(2, 1),  # Optional[Tuple[int, int]]
                pool2_kernel_size=(1, 1),  # Tuple[int, int]
                pool2_stride=(1, 1),  # Optional[Tuple[int, int]]
                out_features=50,  # int
                output_shape=[10, 25, 50],  # Tuple[int, int, int]
                in_sequence_mask=torch.Tensor(10 * [25 * [True] + 25 * [False]]).bool(),  # torch.Tensor
                out_sequence_mask=torch.Tensor(10 * [13 * [True] + 12 * [False]]).bool(),  # torch.Tensor
            ),
            VGG4LayerActTestParams(  # idx 2
                batch=10,  # int
                time=50,  # int
                features=50,  # int
                conv1_channels=1,  # int
                conv2_channels=1,  # int
                conv3_channels=1,  # int
                conv4_channels=1,  # int
                pool1_kernel_size=(1, 1),  # Tuple[int, int]
                pool1_stride=(1, 1),  # Optional[Tuple[int, int]]
                pool2_kernel_size=(2, 1),  # Tuple[int, int]
                pool2_stride=(2, 1),  # Optional[Tuple[int, int]]
                out_features=50,  # int
                output_shape=[10, 25, 50],  # Tuple[int, int, int]
                in_sequence_mask=torch.Tensor(10 * [25 * [True] + 25 * [False]]).bool(),  # torch.Tensor
                out_sequence_mask=torch.Tensor(10 * [13 * [True] + 12 * [False]]).bool(),  # torch.Tensor
            ),
            VGG4LayerActTestParams(  # idx 3
                batch=10,  # int
                time=50,  # int
                features=50,  # int
                conv1_channels=1,  # int
                conv2_channels=1,  # int
                conv3_channels=1,  # int
                conv4_channels=1,  # int
                pool1_kernel_size=(2, 1),  # Tuple[int, int]
                pool1_stride=(2, 1),  # Optional[Tuple[int, int]]
                pool2_kernel_size=(2, 1),  # Tuple[int, int]
                pool2_stride=(2, 1),  # Optional[Tuple[int, int]]
                out_features=50,  # int
                output_shape=[10, 12, 50],  # Tuple[int, int, int]
                in_sequence_mask=torch.Tensor(10 * [25 * [True] + 25 * [False]]).bool(),  # torch.Tensor
                out_sequence_mask=torch.Tensor(10 * [7 * [True] + 5 * [False]]).bool(),  # torch.Tensor
            ),
            VGG4LayerActTestParams(  # idx 4
                batch=10,  # int
                time=50,  # int
                features=50,  # int
                conv1_channels=2,  # int
                conv2_channels=2,  # int
                conv3_channels=2,  # int
                conv4_channels=2,  # int
                pool1_kernel_size=(1, 1),  # Tuple[int, int]
                pool1_stride=(1, 1),  # Optional[Tuple[int, int]]
                pool2_kernel_size=(1, 1),  # Tuple[int, int]
                pool2_stride=(1, 1),  # Optional[Tuple[int, int]]
                out_features=100,  # int
                output_shape=[10, 50, 100],  # Tuple[int, int, int]
                in_sequence_mask=torch.Tensor(10 * [25 * [True] + 25 * [False]]).bool(),  # torch.Tensor
                out_sequence_mask=torch.Tensor(10 * [25 * [True] + 25 * [False]]).bool(),  # torch.Tensor
            ),
            VGG4LayerActTestParams(  # idx 5
                batch=10,  # int
                time=50,  # int
                features=50,  # int
                conv1_channels=3,  # int
                conv2_channels=3,  # int
                conv3_channels=3,  # int
                conv4_channels=3,  # int
                pool1_kernel_size=(1, 1),  # Tuple[int, int]
                pool1_stride=(1, 1),  # Optional[Tuple[int, int]]
                pool2_kernel_size=(1, 1),  # Tuple[int, int]
                pool2_stride=(1, 1),  # Optional[Tuple[int, int]]
                out_features=150,  # int
                output_shape=[10, 50, 150],  # Tuple[int, int, int]
                in_sequence_mask=torch.Tensor(10 * [25 * [True] + 25 * [False]]).bool(),  # torch.Tensor
                out_sequence_mask=torch.Tensor(10 * [25 * [True] + 25 * [False]]).bool(),  # torch.Tensor
            ),
            VGG4LayerActTestParams(  # idx 6
                batch=10,  # int
                time=50,  # int
                features=50,  # int
                conv1_channels=4,  # int
                conv2_channels=4,  # int
                conv3_channels=4,  # int
                conv4_channels=4,  # int
                pool1_kernel_size=(1, 1),  # Tuple[int, int]
                pool1_stride=(1, 1),  # Optional[Tuple[int, int]]
                pool2_kernel_size=(1, 1),  # Tuple[int, int]
                pool2_stride=(1, 1),  # Optional[Tuple[int, int]]
                out_features=200,  # int
                output_shape=[10, 50, 200],  # Tuple[int, int, int]
                in_sequence_mask=torch.Tensor(10 * [25 * [True] + 25 * [False]]).bool(),  # torch.Tensor
                out_sequence_mask=torch.Tensor(10 * [25 * [True] + 25 * [False]]).bool(),  # torch.Tensor
            ),
            VGG4LayerActTestParams(  # idx 7
                batch=10,  # int
                time=50,  # int
                features=50,  # int
                conv1_channels=32,  # int
                conv2_channels=32,  # int
                conv3_channels=64,  # int
                conv4_channels=64,  # int
                pool1_kernel_size=(1, 1),  # Tuple[int, int]
                pool1_stride=(1, 1),  # Optional[Tuple[int, int]]
                pool2_kernel_size=(1, 1),  # Tuple[int, int]
                pool2_stride=(1, 1),  # Optional[Tuple[int, int]]
                out_features=3200,  # int
                output_shape=[10, 50, 3200],  # Tuple[int, int, int]
                in_sequence_mask=torch.Tensor(10 * [25 * [True] + 25 * [False]]).bool(),  # torch.Tensor
                out_sequence_mask=torch.Tensor(10 * [25 * [True] + 25 * [False]]).bool(),  # torch.Tensor
            ),
            VGG4LayerActTestParams(  # idx 8
                batch=10,  # int
                time=50,  # int
                features=50,  # int
                conv1_channels=32,  # int
                conv2_channels=32,  # int
                conv3_channels=64,  # int
                conv4_channels=64,  # int
                pool1_kernel_size=(1, 1),  # Tuple[int, int]
                pool1_stride=(1, 1),  # Optional[Tuple[int, int]]
                pool2_kernel_size=(1, 1),  # Tuple[int, int]
                pool2_stride=(1, 1),  # Optional[Tuple[int, int]]
                out_features=50,  # int
                output_shape=[10, 50, 50],  # Tuple[int, int, int]
                in_sequence_mask=torch.Tensor(10 * [25 * [True] + 25 * [False]]).bool(),  # torch.Tensor
                out_sequence_mask=torch.Tensor(10 * [25 * [True] + 25 * [False]]).bool(),  # torch.Tensor
            ),
            VGG4LayerActTestParams(  # idx 9
                batch=10,  # int
                time=50,  # int
                features=60,  # int
                conv1_channels=4,  # int
                conv2_channels=4,  # int
                conv3_channels=4,  # int
                conv4_channels=4,  # int
                pool1_kernel_size=(1, 2),  # Tuple[int, int]
                pool1_stride=(1, 2),  # Optional[Tuple[int, int]]
                pool2_kernel_size=(1, 2),  # Tuple[int, int]
                pool2_stride=(1, 2),  # Optional[Tuple[int, int]]
                out_features=60,  # int
                output_shape=[10, 50, 60],  # Tuple[int, int, int]
                in_sequence_mask=torch.Tensor(10 * [25 * [True] + 25 * [False]]).bool(),  # torch.Tensor
                out_sequence_mask=torch.Tensor(10 * [25 * [True] + 25 * [False]]).bool(),  # torch.Tensor
            ),
            VGG4LayerActTestParams(  # idx 10
                batch=10,  # int
                time=50,  # int
                features=50,  # int
                conv1_channels=3,  # int
                conv2_channels=3,  # int
                conv3_channels=3,  # int
                conv4_channels=3,  # int
                pool1_kernel_size=(2, 3),  # Tuple[int, int]
                pool1_stride=(2, 3),  # Optional[Tuple[int, int]]
                pool2_kernel_size=(1, 2),  # Tuple[int, int]
                pool2_stride=(1, 2),  # Optional[Tuple[int, int]]
                out_features=24,  # int
                output_shape=[10, 25, 24],  # Tuple[int, int, int]
                in_sequence_mask=torch.Tensor(10 * [25 * [True] + 25 * [False]]).bool(),  # torch.Tensor
                out_sequence_mask=torch.Tensor(10 * [13 * [True] + 12 * [False]]).bool(),  # torch.Tensor
            ),
            VGG4LayerActTestParams(  # idx 11
                batch=10,  # int
                time=50,  # int
                features=70,  # int
                conv1_channels=4,  # int
                conv2_channels=4,  # int
                conv3_channels=4,  # int
                conv4_channels=4,  # int
                pool1_kernel_size=(1, 2),  # Tuple[int, int]
                pool1_stride=(1, 2),  # Optional[Tuple[int, int]]
                pool2_kernel_size=(1, 2),  # Tuple[int, int]
                pool2_stride=(1, 2),  # Optional[Tuple[int, int]]
                out_features=68,  # int
                output_shape=[10, 50, 68],  # Tuple[int, int, int]
                in_sequence_mask=torch.Tensor(10 * [25 * [True] + 25 * [False]]).bool(),  # torch.Tensor
                out_sequence_mask=torch.Tensor(10 * [25 * [True] + 25 * [False]]).bool(),  # torch.Tensor
            ),
            VGG4LayerActTestParams(  # idx 12
                batch=10,  # int
                time=50,  # int
                features=50,  # int
                conv1_channels=32,  # int
                conv2_channels=32,  # int
                conv3_channels=64,  # int
                conv4_channels=64,  # int
                pool1_kernel_size=(4, 3),  # Tuple[int, int]
                pool1_stride=(4, 3),  # Optional[Tuple[int, int]]
                pool2_kernel_size=(1, 3),  # Tuple[int, int]
                pool2_stride=(1, 3),  # Optional[Tuple[int, int]]
                out_features=320,  # int
                output_shape=[10, 12, 320],  # Tuple[int, int, int]
                in_sequence_mask=torch.Tensor(10 * [25 * [True] + 25 * [False]]).bool(),  # torch.Tensor
                out_sequence_mask=torch.Tensor(10 * [7 * [True] + 5 * [False]]).bool(),  # torch.Tensor
            ),
        ]
    ):
        print(idx)
        shape, seq_mask = get_output_shape(test_params)
        print(seq_mask[0])
        print(test_params.out_sequence_mask[0])
        assert list(shape) == test_params.output_shape, (type(shape), type(test_params.output_shape))
        assert torch.equal(seq_mask, test_params.out_sequence_mask), (
            seq_mask.shape,
            test_params.out_sequence_mask.shape,
        )



from __future__ import annotations

import torch

from i6_models.parts.mhsa import MultiheadSelfAttentionV1, MultiheadSelfAttentionV1Config


def test_MultiheadSelfAttentionV1():
    """
    Test the functionality of the MultiheadSelfAttentionV1 module.
    """

    def get_output_shape(input_shape, cfg, key_padding_mask=None, need_weights=True):
        input_tensor = torch.randn(input_shape)
        mhsa = MultiheadSelfAttentionV1(cfg)
        output, weights = mhsa(input_tensor, key_padding_mask)
        return output.shape, weights.shape

    cfg = MultiheadSelfAttentionV1Config(input_dim=32, num_att_heads=8, att_weights_dropout=0.2, dropout=0.3)
    input_shape = [4, 15, 32]  # B,T,F

    key_padding_mask = torch.randint(0, 2, (input_shape[0], input_shape[1])) > 0

    assert get_output_shape(input_shape, cfg, key_padding_mask) == (torch.Size([4, 15, 32]), torch.Size([4, 8, 15, 15]))


def test_ComparisonMHSAV1Torch():
    """
    Compares the output of the MultiheadSelfAttentionV1 module with the output of the torch.nn.MultiheadAttention module.
    """
    cfg = MultiheadSelfAttentionV1Config(input_dim=32, num_att_heads=8, att_weights_dropout=0, dropout=0)
    torch_mhsa = torch.nn.MultiheadAttention(cfg.input_dim, cfg.num_att_heads, dropout=0, batch_first=True)
    torch_mhsa.eval()

    mhsav1 = MultiheadSelfAttentionV1(cfg)
    mhsav1.eval()

    in_proj_weight = torch_mhsa.in_proj_weight
    in_proj_bias = torch_mhsa.in_proj_bias

    out_proj_weight = torch_mhsa.out_proj.weight
    out_proj_bias = torch_mhsa.out_proj.bias

    mhsav1.in_proj.weight = in_proj_weight
    mhsav1.in_proj.bias = in_proj_bias
    mhsav1.out_proj.weight = out_proj_weight
    mhsav1.out_proj.bias = out_proj_bias

    input_shape = [4, 15, 32]  # B,T,F
    input_tensor = torch.randn(input_shape)

    key_padding_mask = torch.randint(0, 2, (input_shape[0], input_shape[1])) > 0

    mhsav1_out, _ = mhsav1(input_tensor, key_padding_mask)
    torch_mhsa_out, _ = torch_mhsa(input_tensor, input_tensor, input_tensor, key_padding_mask)

    assert torch.allclose(mhsav1_out, torch_mhsa_out, atol=1e-08)



from typing import Tuple, Union

import torch
from torch import nn
from torch.nn import functional


def get_same_padding(input_size: Union[int, Tuple[int, ...]]) -> Union[int, Tuple[int, ...]]:
    """
    get padding in order to not reduce the time dimension

    :param input_size:
    :return:
    """
    if isinstance(input_size, int):
        return (input_size - 1) // 2
    elif isinstance(input_size, tuple):
        return tuple((s - 1) // 2 for s in input_size)
    else:
        raise TypeError(f"unexpected size type {type(input_size)}")


def apply_same_padding(x: torch.Tensor, kernel_size: Union[int, Tuple[int, ...]], **kwargs) -> torch.Tensor:
    """
    Pad tensor almost symmetrically in one or more dimensions in order to not reduce time dimension
    when applying convolution with the given kernel. As opposed to the standard padding parameter
    this also handles even kernel sizes.

    :param x:
    :param kernel_size: kernel size of the convolution for which the tensor is padded
    :param kwargs: keyword args passed to functional.pad
    :return: padded tensor
    """
    if isinstance(kernel_size, int):
        h = (kernel_size - 1) // 2
        return functional.pad(x, (h, kernel_size - 1 - h), **kwargs)
    elif isinstance(kernel_size, tuple):
        paddings = ()
        for k in reversed(kernel_size):  # padding function starts with last dim
            h = (k - 1) // 2
            paddings += (h, k - 1 - h)
        return functional.pad(x, paddings, **kwargs)
    else:
        raise TypeError(f"Unexpected size type {type(kernel_size)}")


def mask_pool(seq_mask: torch.Tensor, *, kernel_size: int, stride: int, padding: int) -> torch.Tensor:
    """
    apply strides to the masking

    :param seq_mask: [B,T]
    :param kernel_size:
    :param stride:
    :param padding:
    :return: [B,T'] using maxpool
    """
    if stride == 1 and 2 * padding == kernel_size - 1:
        return seq_mask

    seq_mask = seq_mask.float()
    seq_mask = torch.unsqueeze(seq_mask, 1)  # [B,1,T]
    seq_mask = nn.functional.max_pool1d(seq_mask, kernel_size, stride, padding)  # [B,1,T']
    seq_mask = torch.squeeze(seq_mask, 1)  # [B,T']
    seq_mask = seq_mask.bool()
    return seq_mask


def calculate_output_dim(in_dim: int, *, filter_size: int, stride: int, padding: int) -> int:
    def ceildiv(a: int, b: int):
        return -(-a // b)

    return ceildiv(in_dim + 2 * padding - (filter_size - 1) * 1, stride)



from __future__ import annotations

__all__ = ["MergerV1Config", "MergerV1"]

from dataclasses import dataclass

import torch
from torch import nn

from i6_models.config import ModelConfiguration


@dataclass
class MergerV1Config(ModelConfiguration):
    """
    Attributes:
        input_dim: input dimension
        kernel_size: kernel size of the depthwise convolution layer
        dropout: dropout probability
    """

    input_dim: int
    kernel_size: int
    dropout: float

    def check_valid(self):
        assert self.kernel_size % 2 == 1, "MergerV1 only supports odd kernel sizes"

    def __post__init__(self):
        super().__post_init__()
        self.check_valid()


class MergerV1(nn.Module):
    def __init__(self, model_cfg: MergerV1Config):
        """
        The merge module to merge the outputs of local extractor and global extractor
        Here we take the best variant from the E-branchformer paper (Fig. 3c), refer to
        https://arxiv.org/abs/2210.00077 for more merge module variants
        """
        super().__init__()

        self.depthwise_conv = nn.Conv1d(
            in_channels=model_cfg.input_dim * 2,
            out_channels=model_cfg.input_dim * 2,
            kernel_size=model_cfg.kernel_size,
            padding=(model_cfg.kernel_size - 1) // 2,
            groups=model_cfg.input_dim * 2,
        )
        self.linear_ff = nn.Linear(in_features=2 * model_cfg.input_dim, out_features=model_cfg.input_dim, bias=True)
        self.dropout = model_cfg.dropout

    def forward(self, x_1: torch.Tensor, x_2: torch.Tensor) -> torch.Tensor:
        x_concat = torch.cat([x_1, x_2], dim=-1)  # [B, T, 2F]
        # conv layers expect shape [B, F, T] so we have to transpose here
        x = x_concat.transpose(1, 2)  # [B, 2F, T]
        x = self.depthwise_conv(x)
        x = x.transpose(1, 2)  # [B, T, 2F]
        x = x + x_concat
        x = self.linear_ff(x)
        x = nn.functional.dropout(x, p=self.dropout, training=self.training)
        return x



from __future__ import annotations

__all__ = [
    "FrontendLayerType",
    "GenericFrontendV1Config",
    "GenericFrontendV1",
]

from dataclasses import dataclass
from enum import Enum, auto
from typing import Callable, Optional, Tuple, Union, Sequence

import torch
from torch import nn

from i6_models.config import ModelConfiguration

from i6_models.parts.frontend.common import get_same_padding, mask_pool, calculate_output_dim


class FrontendLayerType(Enum):
    Conv2d = auto()
    Pool2d = auto()
    Activation = auto()


@dataclass
class GenericFrontendV1Config(ModelConfiguration):
    """
    Attributes:
        in_features: number of input features to module
        layer_ordering: the ordering of the front end layer sequences, the ordering element must be selected from FrontendLayerType
            e.g. the ordering of VGG4LayerActFrontendV1 would be [FrontendLayerType.Conv2d, FrontendLayerType.Activation,
            FrontendLayerType.Pool2d, FrontendLayerType.Conv2d, FrontendLayerType.Conv2d, FrontendLayerType.Activation,
            FrontendLayerType.Pool2d]
        conv_kernel_sizes: kernel sizes for each conv layer
        conv_strides: stride sizes for each conv layer
        conv_paddings: paddings sizes for each conv layer
        conv_out_dims: number of out channels for each conv layer
        pool_kernel_sizes: kernel sizes for each pool layer
        pool_strides: stride sizes for each pool layer
        pool_paddings: padding sizes for each pool layer
        activations: activation functions
        out_features: output size of the final linear layer
    """

    in_features: int
    layer_ordering: Sequence[FrontendLayerType]
    conv_kernel_sizes: Optional[Sequence[Tuple[int, int]]]
    conv_strides: Optional[Sequence[Tuple[int, int]]]
    conv_paddings: Optional[Sequence[Tuple[int, int]]]
    conv_out_dims: Optional[Sequence[int]]
    pool_kernel_sizes: Optional[Sequence[Tuple[int, int]]]
    pool_strides: Optional[Sequence[Tuple[int, int]]]
    pool_paddings: Optional[Sequence[Tuple[int, int]]]
    activations: Optional[Sequence[Union[nn.Module, Callable[[torch.Tensor], torch.Tensor]]]]
    out_features: int

    def check_valid(self):
        num_convs = 0 if self.conv_kernel_sizes is None else len(self.conv_kernel_sizes)
        num_pools = 0 if self.pool_kernel_sizes is None else len(self.pool_kernel_sizes)
        num_activations = 0 if self.activations is None else len(self.activations)

        assert num_convs == self.layer_ordering.count(
            FrontendLayerType.Conv2d
        ), "Number of convolution layers mismatch!"
        assert num_activations == self.layer_ordering.count(
            FrontendLayerType.Activation
        ), "Number of activation layers mismatch!"
        assert num_pools == self.layer_ordering.count(FrontendLayerType.Pool2d), "Number of pooling layers mismatch!"

        if self.conv_strides is not None:
            assert len(self.conv_strides) == num_convs, "Please specify stride for each convolution layer!"
        if self.conv_paddings is not None:
            assert len(self.conv_paddings) == num_convs, "Please specify padding for each convolution layer!"
        if num_convs != 0:
            assert (
                len(self.conv_out_dims) == num_convs
            ), "Please specify the number of channels for each convolution layer!"

        if self.pool_strides is not None:
            assert len(self.pool_strides) == num_pools, "Please specify stride for each pooling layer!"
        if self.pool_paddings is not None:
            assert len(self.pool_paddings) == num_pools, "Please specify padding for each pooling layer!"

        assert len(self.layer_ordering) == num_convs + num_pools + num_activations, "Number of total layers mismatch!"

        for kernel_sizes in filter(None, [self.conv_kernel_sizes, self.pool_kernel_sizes]):
            for kernel_size in kernel_sizes:
                assert all(k % 2 for k in kernel_size), "ConformerVGGFrontendV1 only supports odd kernel sizes"

    def __post__init__(self):
        super().__post_init__()
        self.check_valid()


class GenericFrontendV1(nn.Module):
    def __init__(self, model_cfg: GenericFrontendV1Config):
        """
        Generic Front-End
        can be used to generate customized frontend by combining convolutional and pooling layers, as well as activation
        functions differently

        To get the ESPnet case, for example Conv2dSubsampling6, use these options
            layer_ordering = [FrontendLayerType.Conv2d, FrontendLayerType.Conv2d]
            conv_kernel_sizes = [3, 5]
            strides = [2, 3]

        To get the i6_models VGG4LayerActFrontendV1, use the options:
            layer_ordering = [FrontendLayerType.Conv2d, FrontendLayerType.Activation, FrontendLayerType.Pool2d,
                FrontendLayerType.Conv2d, FrontendLayerType.Conv2d, FrontendLayerType.Activation, FrontendLayerType.Pool2d]
            conv_kernel_sizes = [3, 3, 3]
            conv_out_dims = [32, 34, 64]
            pool_kernel_sizes = [3, 3]
            pool_strides = [2, 2]
            activations = [torch.nn.ReLU(), torch.nn.ReLU()]
        """
        super().__init__()

        model_cfg.check_valid()

        self.cfg = model_cfg

        self.frontend_layers = nn.ModuleList([])

        conv_layer_index = 0
        pool_layer_index = 0
        activation_layer_index = 0
        last_channel_dim = 1
        last_feat_dim = model_cfg.in_features
        for layer_type in model_cfg.layer_ordering:
            if layer_type == FrontendLayerType.Conv2d:
                conv_out_dim = model_cfg.conv_out_dims[conv_layer_index]
                conv_kernel_size = model_cfg.conv_kernel_sizes[conv_layer_index]
                conv_stride = 1 if model_cfg.conv_strides is None else model_cfg.conv_strides[conv_layer_index]
                conv_padding = (
                    get_same_padding(conv_kernel_size)
                    if model_cfg.conv_paddings is None
                    else model_cfg.conv_paddings[conv_layer_index]
                )

                self.frontend_layers.append(
                    nn.Conv2d(
                        in_channels=last_channel_dim,
                        out_channels=conv_out_dim,
                        kernel_size=conv_kernel_size,
                        stride=conv_stride,
                        padding=conv_padding,
                    )
                )

                last_channel_dim = conv_out_dim
                last_feat_dim = calculate_output_dim(
                    in_dim=last_feat_dim,
                    filter_size=conv_kernel_size[1],
                    stride=conv_stride[1],
                    padding=conv_padding[1],
                )
                conv_layer_index += 1

            elif layer_type == FrontendLayerType.Pool2d:
                pool_stride = None if model_cfg.pool_strides is None else model_cfg.pool_strides[pool_layer_index]
                pool_kernel_size = model_cfg.pool_kernel_sizes[pool_layer_index]
                pool_padding = (
                    get_same_padding(pool_kernel_size)
                    if model_cfg.pool_paddings is None
                    else model_cfg.pool_paddings[pool_layer_index]
                )

                self.frontend_layers.append(
                    nn.MaxPool2d(
                        kernel_size=pool_kernel_size,
                        stride=pool_stride,
                        padding=pool_padding,
                    )
                )
                last_feat_dim = calculate_output_dim(
                    in_dim=last_feat_dim,
                    filter_size=pool_kernel_size[1],
                    stride=pool_stride[1] or pool_kernel_size[1],
                    padding=pool_padding[1],
                )
                pool_layer_index += 1

            elif layer_type == FrontendLayerType.Activation:
                self.frontend_layers.append(model_cfg.activations[activation_layer_index])
                activation_layer_index += 1
            else:
                raise NotImplementedError

            self.linear = nn.Linear(
                in_features=last_feat_dim * last_channel_dim,
                out_features=model_cfg.out_features,
                bias=True,
            )

    def forward(self, tensor: torch.Tensor, sequence_mask: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        assert tensor.shape[-1] == self.cfg.in_features, f"shape {tensor.shape} vs in features {self.cfg.in_features}"
        # and add a dim
        tensor = tensor[:, None, :, :]  # [B,C=1,T,F]

        for layer in self.frontend_layers:
            tensor = layer(tensor)

            if isinstance(layer, nn.Conv2d) or isinstance(layer, nn.MaxPool2d):
                sequence_mask = mask_pool(
                    sequence_mask,
                    kernel_size=layer.kernel_size[0],
                    stride=layer.stride[0],
                    padding=layer.padding[0],
                )

        tensor = torch.transpose(tensor, 1, 2)  # transpose to [B,T",C,F"]
        tensor = torch.flatten(tensor, start_dim=2, end_dim=-1)  # [B,T",C*F"]

        tensor = self.linear(tensor)

        return tensor, sequence_mask



__all__ = ["MixupConfig", "Mixup"]

from dataclasses import dataclass
from typing import Optional

import torch

from i6_models.config import ModelConfiguration


@dataclass
class MixupConfig(ModelConfiguration):
    """
    Attributes:
        feature_dim: feature dimension
        apply_prob: probability to apply mixup at all
        lambda_min: minimum of lambda_ value (suppose input = input + lambda_ * mixup_values)
        lambda_max: maximum of lambda_ value (suppose input = input + lambda_ * mixup_values)
        max_num_mix: maximum number of mixups (random int in [1, max_num_mix])
        buffer_size: number of frames.
    """

    feature_dim: int
    apply_prob: float
    max_num_mix: int
    lambda_min: float
    lambda_max: float
    buffer_size: int = 1_000_000


class FeatureBuffer(torch.nn.Module):
    """
    The FeatureBuffer saves the feature from previous timeframes
    after the buffer is full, the buffer will not be updated
    """

    def __init__(self, *, buffer_size: int, feature_dim: int):
        super().__init__()
        self.filled = False
        self.pos = 0
        self.buffer_size = buffer_size
        self.register_buffer("cache", torch.zeros((self.buffer_size, feature_dim)))

    def append(self, tensor: torch.Tensor):
        """
        :param tensor: [T', F]
        """
        t_dim = tensor.shape[0]
        if t_dim > self.buffer_size:
            tensor = tensor[: self.buffer_size]
            t_dim = self.buffer_size
        delta_pos = min(self.buffer_size - self.pos, t_dim)
        end_pos = self.pos + delta_pos
        self.cache[self.pos : end_pos] = tensor[:delta_pos]
        self.pos = end_pos

        if end_pos == self.buffer_size:
            self.filled = True
            end_pos = t_dim - delta_pos
            self.cache[:end_pos] = tensor[delta_pos:]
            self.pos = end_pos

    def get_random(self, b_dim: int, t_dim: int, max_num_mixup: int, n_mask: torch.Tensor) -> Optional[torch.Tensor]:
        """
        :param n_mask: [B, M]
        :returnn tensor as # [T, M', F]
        """
        if not self.filled and self.pos == 0:
            return None
        else:
            end_idx = self.buffer_size if self.filled else self.pos
            max_end_idx = end_idx - t_dim

            start_indicies = torch.randint(
                high=max_end_idx, size=(b_dim, max_num_mixup)
            )  # [B, M] (M denotes maximum of num_mixup over the batch)
            start_indicies_flat = torch.masked_select(
                start_indicies, n_mask
            )  # [M'] (M' denotes sum of num_mixup over the batch)

            idx = torch.arange(t_dim)  # [T]
            idx = idx[:, None] + start_indicies_flat[None, :]  # [T, M']
            mixup_values = self.cache[idx]  # [T, M', F]
            return mixup_values


class Mixup(torch.nn.Module):
    """
    Implement the Mixup data augmentation method
    C.f. https://arxiv.org/abs/1710.09412

    The code is partly based on Albert's implementation in returnn front-end
    C.f. https://github.com/rwth-i6/i6_experiments/blob/main/users/zeyer/returnn/models/rf_mixup.py
    """

    def __init__(self, *, cfg: MixupConfig):
        super().__init__()
        self.apply_prob = cfg.apply_prob
        self.lambda_min = cfg.lambda_min
        self.lambda_max = cfg.lambda_max
        self.max_num_mix = cfg.max_num_mix
        self.buffer_size = cfg.buffer_size
        self.feature_dim = cfg.feature_dim

        self.feature_buffer = FeatureBuffer(buffer_size=self.buffer_size, feature_dim=self.feature_dim)

    def forward(self, input: torch.Tensor, /, sequence_mask: torch.Tensor) -> torch.Tensor:
        """
        :param input: [B, T, F]
        :param sequence_mask: [B, T]
        :return tensor as [B, T, F]
        """
        assert (
            input.size()[-1] == self.feature_dim
        ), "the given feature dimension does not match input feature dimension"
        input_ = self._maybe_apply_mixup(input, sequence_mask)
        self._append_to_buffer(input, sequence_mask)
        return input_

    def _append_to_buffer(self, input: torch.Tensor, sequence_mask: torch.Tensor):
        """
        :param input: [B, T, F]
        :param sequence_mask: [B, T]
        """
        _, _, f_dim = input.size()
        # mask out the padded frames before append it to feature_buffer
        input_flat = torch.masked_select(input, sequence_mask.unsqueeze(-1)).view(-1, f_dim)  # [B*T, F]
        self.feature_buffer.append(input_flat)

    def _maybe_apply_mixup(self, input: torch.Tensor, sequence_mask: torch.Tensor) -> torch.Tensor:
        """
        :param input: [B, T, F]
        :param sequence_mask: [B, T]
        :return tensor as [B,T,F]
        """
        if torch.randn(()) >= self.apply_prob:
            return input

        b_dim, t_dim, f_dim = input.size()  # [B, T, F]

        max_end_idx = self.feature_buffer.buffer_size if self.feature_buffer.filled else self.feature_buffer.pos
        if t_dim > max_end_idx:
            return input

        num_mixup = torch.randint(low=1, high=self.max_num_mix + 1, size=(b_dim,))  # [B]
        max_num_mixup = num_mixup.max()  # scalar, =M (M denotes maximum of num_mixup)

        row_vector = torch.arange(0, max_num_mixup, 1)  # [M]
        n_mask = row_vector[None, :] < num_mixup[:, None]  # [B, M]

        mixup_values = self.feature_buffer.get_random(b_dim, t_dim, max_num_mixup, n_mask).to(
            input.device
        )  # [T, M', F] (M' denotes sum of num_mixup over the batch)

        lambda_ = torch.FloatTensor(b_dim, max(num_mixup)).uniform_(self.lambda_min, self.lambda_max)  # [B, M]
        mixup_scales = torch.FloatTensor(b_dim, max(num_mixup)).uniform_(0.001, 1)  # [B, M]

        mixup_scales *= lambda_ / torch.unsqueeze(torch.sum(mixup_scales * n_mask, axis=1), dim=-1)  # [B, M]
        mixup_scales_flat = torch.masked_select(mixup_scales, n_mask)  # [M']
        mixup_values = torch.einsum("ijk,j->ijk", mixup_values, mixup_scales_flat.to(mixup_values.device))  # [T, M', F]

        idx_b = torch.arange(0, b_dim, 1)  # [B]
        idx_b = torch.masked_select(idx_b.unsqueeze(-1).expand(n_mask.size()), n_mask)  # [M']
        ones = torch.ones(mixup_values.size())  # [T, M', F]
        idx_b = torch.einsum("ijk,j->ijk", ones, idx_b).to(mixup_values.device, dtype=torch.long)  # [T, M', F]

        mixup_value = torch.scatter_add(
            torch.zeros(t_dim, b_dim, f_dim).to(mixup_values.device), 1, idx_b, mixup_values
        )  # [T, B, F]
        mixup_value = mixup_value.transpose(0, 1)  # [B, T, F]

        # mask out padded frames
        mixup_value = torch.mul(mixup_value, sequence_mask.unsqueeze(-1))  # [B, T, F]

        input = input + mixup_value  # [B, T, F]
        return input



from __future__ import annotations

__all__ = ["ConformerConvolutionV1", "ConformerConvolutionV1Config"]

from dataclasses import dataclass
from copy import deepcopy

import torch
from torch import nn
from i6_models.config import ModelConfiguration
from typing import Callable, Union


@dataclass
class ConformerConvolutionV1Config(ModelConfiguration):
    """
    Attributes:
        channels: number of channels for conv layers
        kernel_size: kernel size of conv layers
        dropout: dropout probability
        activation: activation function applied after normalization
        norm: normalization layer with input of shape [N,C,T]
    """

    channels: int
    kernel_size: int
    dropout: float
    activation: Union[nn.Module, Callable[[torch.Tensor], torch.Tensor]]
    norm: Union[nn.Module, Callable[[torch.Tensor], torch.Tensor]]

    def check_valid(self):
        assert self.kernel_size % 2 == 1, "ConformerConvolutionV1 only supports odd kernel sizes"

    def __post_init__(self):
        super().__post_init__()
        self.check_valid()


class ConformerConvolutionV1(nn.Module):
    """
    Conformer convolution module.
    see also: https://github.com/espnet/espnet/blob/713e784c0815ebba2053131307db5f00af5159ea/espnet/nets/pytorch_backend/conformer/convolution.py#L13

    Uses explicit padding for ONNX exportability, see:
    https://github.com/pytorch/pytorch/issues/68880
    """

    def __init__(self, model_cfg: ConformerConvolutionV1Config):
        """
        :param model_cfg: model configuration for this module
        """
        super().__init__()
        model_cfg.check_valid()
        self.pointwise_conv1 = nn.Linear(in_features=model_cfg.channels, out_features=2 * model_cfg.channels)
        self.depthwise_conv = nn.Conv1d(
            in_channels=model_cfg.channels,
            out_channels=model_cfg.channels,
            kernel_size=model_cfg.kernel_size,
            padding=(model_cfg.kernel_size - 1) // 2,
            groups=model_cfg.channels,
        )
        self.pointwise_conv2 = nn.Linear(in_features=model_cfg.channels, out_features=model_cfg.channels)
        self.layer_norm = nn.LayerNorm(model_cfg.channels)
        self.norm = deepcopy(model_cfg.norm)
        self.dropout = nn.Dropout(model_cfg.dropout)
        self.activation = model_cfg.activation

    def forward(self, tensor: torch.Tensor) -> torch.Tensor:
        """
        :param tensor: input tensor of shape [B,T,F]
        :return: torch.Tensor of shape [B,T,F]
        """
        tensor = self.layer_norm(tensor)
        tensor = self.pointwise_conv1(tensor)  # [B,T,2F]
        tensor = nn.functional.glu(tensor, dim=-1)  # [B,T,F]

        # conv layers expect shape [B,F,T] so we have to transpose here
        tensor = tensor.transpose(1, 2)  # [B,F,T]
        tensor = self.depthwise_conv(tensor)

        tensor = self.norm(tensor)
        tensor = tensor.transpose(1, 2)  # transpose back to [B,T,F]

        tensor = self.activation(tensor)
        tensor = self.pointwise_conv2(tensor)

        return self.dropout(tensor)



from itertools import product

import torch
from torch import nn

from i6_models.parts.frontend.common import apply_same_padding, get_same_padding


def test_output_shape():
    # test for even and odd dim
    last_dim = 101
    pre_last_dim = 100

    iff = lambda x, y: x and y or not x and not y  # x <=> y
    strided_dim = lambda d, s: (d - 1) // s + 1  # expected out dimension for strided conv

    # `get_same_padding` seems to work for some stride > 1
    for kernel in product(range(1, 21), repeat=2):
        conv = nn.Conv2d(1, 1, kernel_size=kernel, stride=(1, 1), padding=get_same_padding(kernel))

        x = torch.randn(1, 1, pre_last_dim, last_dim)

        out = conv(x)

        # we expect `get_same_padding` to only cover odd kernel sizes
        assert all(
            iff(out_dim == in_dim, k % 2 == 1) for in_dim, out_dim, k in zip(x.shape[2:], out.shape[2:], kernel)
        ), f"Failed for {x.shape=}, {out.shape=}, {kernel=} and stride=1"

    for kernel, stride in product(product(range(1, 21), repeat=2), range(1, 7)):
        conv = nn.Conv2d(1, 1, kernel_size=kernel, stride=(1, stride))

        x = torch.randn(1, 1, pre_last_dim, last_dim)
        x_padded = apply_same_padding(x, kernel)

        out = conv(x_padded)

        # correct out dimensions for all possible kernel sizes and strides
        assert all(
            out_dim == strided_dim(in_dim, s)
            for in_dim, out_dim, k, s in zip(x.shape[2:], out.shape[2:], kernel, (1, stride))
        ), f"Failed for {x.shape=}, {out.shape=}, {kernel=} and {stride=}"



import onnxruntime as ort
import tempfile
import torch
from torch.onnx import export as export_onnx

from i6_models.parts.blstm import BlstmEncoderV1, BlstmEncoderV1Config


def test_blstm_onnx_export():
    with torch.no_grad(), tempfile.NamedTemporaryFile() as f:
        config = BlstmEncoderV1Config(num_layers=4, input_dim=50, hidden_dim=128, dropout=0.1, enforce_sorted=True)
        model = BlstmEncoderV1(config=config)
        scripted_model = torch.jit.optimize_for_inference(torch.jit.script(model.eval()))

        dummy_data = torch.randn(3, 30, 50)
        dummy_data_len = torch.IntTensor([30, 20, 15])
        dummy_data_len_2 = torch.IntTensor([30, 15, 10])

        outputs_normal = model(dummy_data, dummy_data_len)
        outputs_scripted = scripted_model(dummy_data, dummy_data_len)
        assert torch.allclose(outputs_normal, outputs_scripted)
        export_onnx(
            scripted_model,
            (dummy_data, dummy_data_len),
            f=f,
            verbose=True,
            input_names=["data", "data_len"],
            output_names=["classes"],
            dynamic_axes={
                # dict value: manually named axes
                "data": {0: "batch", 1: "time"},
                "data_len": {0: "batch"},
                "classes": {0: "batch", 1: "time"},
            },
        )
        session = ort.InferenceSession(f.name, providers=["CPUExecutionProvider"])
        outputs_onnx = torch.FloatTensor(
            session.run(None, {"data": dummy_data.numpy(), "data_len": dummy_data_len.numpy()})[0]
        )
        outputs_onnx_other = torch.FloatTensor(
            session.run(None, {"data": dummy_data.numpy(), "data_len": dummy_data_len_2.numpy()})[0]
        )
        # The default 1e-8 was slightly too strong
        assert torch.allclose(outputs_normal, outputs_onnx, atol=1e-6)
        # check that for different lengths we really get a different result
        assert not torch.allclose(outputs_normal, outputs_onnx_other, atol=1e-6)

        # check with different batching and max size
        outputs_onnx_diff_batch = torch.FloatTensor(
            session.run(
                None,
                {
                    "data": dummy_data[(1, 2), :20, :].numpy(),
                    "data_len": dummy_data_len[
                        (1, 2),
                    ].numpy(),
                },
            )[0]
        )
        assert torch.allclose(outputs_normal[2, :20], outputs_onnx_diff_batch[1], atol=1e-6)
